--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

module(..., package.seeall)
require'Class'
require'Vector2d'

local CameraMetatable = Class(Camera)

-- Creates a new Camera.
-- @param position Vector2d describing the camera's position.
-- @param zoom Number describing the zoom level [1].
-- @param rotation Number describing the camera's rotation in radians [0].
function Camera.new(position, zoom, rotation)
	local camera = {
		position = position:copy(),
		zoom = zoom or 1,
		rotation = rotation or 0,
	}

	setmetatable(camera, CameraMetatable)
	return camera
end

-- Rotates the camera by the given angle.
-- @param self The camera.
-- @param phi Angle to rotate the camera in radians.
-- @return The camera.
function rotate(self, phi)
	self.rotation = self.rotation + phi
	return self
end

-- Moves the camera by the measurement described in the Vector2d.
-- @param self The camera.
-- @param vector Vector2d describing the movement.
function move(self, vector)
	self.position = self.position:add(vector)
	return self
end

-- Moves the camera to the absolute world position.
-- @param self The camera.
-- @param vector Vector2d describing the position.
function moveAbsolute(self, vector)
	self.position = vector:copy()
	return self
end

-- Attaches the camera.
-- @param self The camera.
function attach(self)
	local scale = self.zoom * 2
	local x, y = love.graphics.getWidth() / scale, love.graphics.getHeight() / scale
	love.graphics.push()
	love.graphics.scale(self.zoom)
	love.graphics.translate(x, y)
	love.graphics.rotate(self.rotation)
	love.graphics.translate(-self.position.x, -self.position.y)
end

-- Detaches the camera.
-- @param self The camera.
function detach(self)
	love.graphics.pop()
end

-- Gets the position of the camera in world coordinates.
-- @param self The camera.
-- @return Camera position in world coordinates.
function getPosition(self)
	return self.position
end

-- Attaches the camera, calls the function and detaches the camera.
-- @param self The camera.
-- @param func Function to call.
function draw(self, func)
	self:attach()
	func()
	self:detach()
end

-- Translates the given world coordinates into camera coordinates.
-- @param self The camera.
-- @param vector Vector2d specifying the world coordinates to translate.
-- @return Vector2d specifying the world coordinates in the camera coordinate space.
function cameraCoordinatesFor(self, vector)
	return vector:subtract(self.position)
		:rotate(self.rotation)
		:scale(self.zoom)
		:add(Vector2d.new(love.graphics.getWidth() / 2, love.graphics.getHeight() / 2))
end

-- Translates the given camera coordinates into world coordinates.
-- @param self The camera.
-- @param vector Vector2d specifying the camera coordinates to translate.
-- @return Vector2d specifying the camera coordinates in the world coordinate space.
function worldCoordinatesFor(self, vector)
	return vector:subtract(Vector2d.new(love.graphics.getWidth() / 2, love.graphics.getHeight() / 2))
		:divide(self.zoom)
		:rotate(-self.rotation)
		:add(self.position)
end

-- Gets the world coordinates of the mouse.
-- @param self The camera
-- @return World coordinates describing the current mouse position.
function mousePosition(self)
	return self:worldCoordinatesFor(Vector2d.new(love.mouse.getPosition()))
end
