--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

module(..., package.seeall)
require'Class'
require'Ship'
require'Vector2d'
require'Thing'
require'Resources'

local ControllableMetatable = Class(Controllable)

local defaultPosition = { x = 0, y = 0 }
local defaultVelocity = {x = 0, y = 0 }
local defaultFriction = 400

function Controllable.new(x, y, acceleration, maxVelocity, friction)
	controllable = {
		position = Vector2d.new(x, y),
		velocity = Vector2d.new(defaultVelocity.x, defaultVelocity.y),
		friction = friction or defaultFriction,
		output = function() end,
		radius = 10, 
		thrusting = false,
		accelerate = false,
		shoot = false,
		bulletImage = { },
		bulletVelocity = Vector2d.new(400, 0),
		timeSinceFired = 0.05,
		timeToFire = 0.05,
		angle = 0,
		targetPosition = Vector2d.new(0, 0),
		targetImage = { },
		ship = Ship.new(acceleration, maxVelocity, 80, 2, 4),
	}
	
	setmetatable(controllable, ControllableMetatable)
	return controllable
end

function Controllable:load( )
	self.image = Resources.Images.Ships.Ship
	self.emissions = Resources.Images.Effects.Flame
	self.bulletImage = Resources.Images.Objects.Bullet
	self.targetImage = Resources.Images.Objects.Crosshairs
	
	self.particle = love.graphics.newParticleSystem(self.emissions, 128)
	self.particle:setEmissionRate          (50)
	self.particle:setLifetime              (0.02)
	self.particle:setParticleLife          (0.4)
	self.particle:setSpread                (0)
	self.particle:setSpeed                 (50, 200)
	self.particle:setGravity               (0)
	self.particle:setRadialAcceleration    (5, 10)
	self.particle:setTangentialAcceleration(5, 10)
	self.particle:setSizes                 (3)
	self.particle:setSizeVariation         (5.7)
	self.particle:setRotation              (0)
	self.particle:setSpin                  (0)
	self.particle:setSpinVariation         (0)
	self.particle:setColors                (20, 20, 25, 240, 155, 155, 155, 10)
end

function Controllable:consumeFuel(units) 
	if (self.fuel - units) < 0 then return false end
	
	self.fuel = self.fuel - units
	return true
end

function Controllable:updateAccelerationParticles(dt)
	if self.thrusting then 
		self.particle:start() 
		self.particle:update(dt)
	end
end

function Controllable:update(thrustVector, dt)
	if not dt then 
		self.output('returning from Controllable:update because dt is nil.')
		return 
	end
	
	self.timeSinceFired = self.timeSinceFired + dt
	self.ship:regenerate(dt)
	
	if not thrustVector.x and not thrustVector.y then return end
	
	local mouseVector = thrustVector:subtract(self.position)
	self.angle = mouseVector:angle()

	if self.accelerate then
		local d = mouseVector:length()
		local normalized = mouseVector:normalize()
		
		local maximumVelocity = self.ship:getMaxVelocity()
		
		if self.ship:consumeFuel(dt) then
			self.thrusting = self.accelerate
			local acceleration = normalized:scale(self.ship:getAcceleration())
			self.velocity = acceleration:scale(dt):add(self.velocity)

			-- if accelerating under the player's power, then cap it at the maximum velocity
			if (self.velocity:length() > maximumVelocity) then
				self.velocity = self.velocity:normalize():scale(maximumVelocity)
			end
		else
			self.thrusting = false
		end
	else
		self.thrusting = false
	end
	
	-- todo: redo the algorithm, not quite fair
	if self.shoot and self.timeToFire < self.timeSinceFired then
		self.timeSinceFired = 0
		self:spawnBullet( )
	end
	
	self.targetPosition = self.bulletVelocity:rotate(self.angle)
						:add(self.velocity)
						:normalize()
						:scale(mouseVector:length())
						:add(self.position)
						:add(Vector2d.new(20, 0):rotate(self.angle))
	
	self.position = self.velocity:scale(dt):add(self.position)
	
	-- TODO bounds function
	main.wrap:ensureWrap(self.position)
	
	self:updateAccelerationParticles(dt)
end

function Controllable:draw( )
	if self.image then
		love.graphics.draw(self.image, 
			self.position.x, 
			self.position.y, 
			self.angle, 
			1, 
			1,
			self.image:getWidth() / 2, 
			self.image:getHeight() / 2)
	
		if self.thrusting then
			local particleDirection = self.angle + math.pi
			local particleSource = Vector2d.new(16, 0):rotate(particleDirection):add(self.position)
			self.particle:setDirection(particleDirection)
			love.graphics.draw(self.particle, particleSource.x, particleSource.y)
		end
	end
		
	--r, g, b, a = love.graphics.getColor()
	--love.graphics.setColor(34, 139, 34, 50)
	--love.graphics.circle('line', self.position.x, self.position.y, 400, 1000)
	--love.graphics.setColor(r, g, b, a)
	-- TODO: should be updated with better logic to obtain size and scale
	love.graphics.draw(self.targetImage, self.targetPosition.x, self.targetPosition.y, 0, 0.5, 0.5, 64, 64)
end

function Controllable:spawnBullet( )
	-- TODO clean this up; constants, etc
	local scale = 0.1
	local position = self.position:add(Vector2d.new(20, 0):rotate(self.angle))
	local boundingShape = main.collider:addRectangle(position.x, position.y, self.bulletImage:getWidth() * scale, self.bulletImage:getHeight() * scale)
	boundingShape:rotate(self.angle + math.pi / 2)
	local bullet = Thing.new(
		Constant.Thing.Bullet,
		function (position, angle) 
			love.graphics.draw(self.bulletImage, position.x, position.y, angle, scale, scale)
		end,
		position,
		self.velocity:add(self.bulletVelocity:rotate(self.angle)),
		self.angle + math.pi / 2,
		boundingShape,
		3000
	)
	
	main.colliderLookup[boundingShape] = bullet
	table.insert(main.entities, bullet)
end

function Controllable:startAcceleration()
	self.accelerate = true
end

function Controllable:stopAcceleration()
	self.accelerate = false
end

function Controllable:startShooting()
	self.shoot = true
end

function Controllable:stopShooting()
	self.shoot = false
end
