--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

module(..., package.seeall)
require'Class'
require'Camera'
require'Resources'
local Constant = require'Constant'

SpaceImages = { 
	StarImages = Resources.Images.Space.StarImages,
	DistantImages = Resources.Images.Space.DistantImages,
	NearImages = Resources.Images.Space.NearImages,
	Orientation = { },
 }

local SpaceMetatable = Class(Space)

function new(size, output)
	local space = wrapAround(size, output)
	setmetatable(space, SpaceMetatable)
	return space
end

function sized(size, output)
	local distantSpace = { }

	for i = 0, size.x * Constant.Space.EighthScale, 512 do
		for j = 0, size.y * Constant.Space.EighthScale, 512 do
			--local orientation = Space.Orientation[math.random(1, 4)]
			table.insert(distantSpace, {
				x = i,
				y = j,
				scale = 1, 
				rotation = math.random(0, 3) * math.pi / 2,
				image = SpaceImages.DistantImages[math.random(1, table.getn(SpaceImages.DistantImages))] 
			})
		end
	end
	
	local nearSpace = { }

	for i = 0, size.x * Constant.Space.QuarterScale, 512 do
		for j = 0, size.y * Constant.Space.QuarterScale, 512 do
			--local orientation = Space.Orientation[math.random(1, 4)]
			table.insert(nearSpace, { 
				x = i, 
				y = j,
				scale = 1, 
				rotation = math.random(0, 3) * math.pi / 2,
				image = SpaceImages.NearImages[math.random(1, table.getn(SpaceImages.NearImages))] 
			})
			
			while math.random() > 0.5 do 
				table.insert(nearSpace, { 
					x = i + math.random(1, 612) - 100, 
					y = j + math.random(1, 612) - 100, 
					scale = 0.2,
					image = SpaceImages.StarImages[math.random(1, table.getn(SpaceImages.StarImages))], 
				}) 
			end
		end
	end
	
	local closeSpace = { }
	
	for i = 0, size.x * Constant.Space.HalfScale, 512 do
		for j = 0, size.y * Constant.Space.HalfScale, 512 do
			while math.random() > 0.8 do 
				table.insert(closeSpace, { 
					x = i + math.random(1, 612) - 100, 
					y = j + math.random(1, 612) - 100, 
					scale = .4, --math.random(0.4, 0.6),
					image = SpaceImages.StarImages[math.random(1, table.getn(SpaceImages.StarImages))], 
				}) 
			end
		end
	end
	
	local space = {
		output = output,
		x = size.x,
		y = size.y,
		distantSpace = distantSpace,
		nearSpace = nearSpace,
		closeSpace = closeSpace,
		distantCamera = Camera.new(Vector2d.new(0, 0)),
		nearCamera = Camera.new(Vector2d.new(0, 0)),
		closeCamera = Camera.new(Vector2d.new(0, 0)),
		position = { x = 0, y = 0 },
	}

	return space
end

function populateLayer(layer, origin, size, scale, tiles, imageFunction, chance)
	for i = origin.x, size.x * scale, 512 do
		for j = origin.y, size.y * scale, 512 do
			table.insert(layer, {
				x = i,
				y = j,
				scale = 1, 
				rotation = math.random(0, 3) * math.pi / 2,
				image = imageFunction() 
			})
		end
	end
end

function addBuffer(source, size)
	local temp = { }
	for _, tile in ipairs(source) do
		table.insert(temp, { 
			x = tile.x + size.x,
			y = tile.y,
			scale = tile.scale,
			rotation = tile.rotation,
			image = tile.image,
		})
		table.insert(temp, {
			x = tile.x,
			y = tile.y + size.x,
			scale = tile.scale,
			rotation = tile.rotation,
			image = tile.image,
		})
		table.insert(temp, {
			x = tile.x + size.x,
			y = tile.y + size.y,
			scale = tile.scale,
			rotation = tile.rotation,
			image = tile.image,
		})
	end
	return temp
end

-- 
--
--
-- @return New Space instance with associated cameras.
function wrapAround( )
	-- 32,768 x 32,768
	local size = Vector2d.new(16384, 16384)
	
	local distantSpace = { }
	
	populateLayer(distantSpace, Vector2d.new(1, 1), size, 
		Constant.Space.EighthScale, SpaceImages.DistantImages, 
		function( ) return SpaceImages.DistantImages[math.random(1, table.getn(SpaceImages.DistantImages))] end, 1)

	for _, tile in ipairs(addBuffer(distantSpace, size:scale(Constant.Space.EighthScale))) do
		table.insert(distantSpace, tile)
	end

	local nearSpace = { }
	
	populateLayer(nearSpace, Vector2d.new(1, 1), size, 
		Constant.Space.QuarterScale, SpaceImages.NearImages, 
		function( ) return SpaceImages.NearImages[math.random(1, table.getn(SpaceImages.NearImages))] end, 1)

	for i = 1, size.x * Constant.Space.QuarterScale, 512 do
		for j = 1, size.y * Constant.Space.QuarterScale, 512 do
			while math.random() > 0.5 do 
				table.insert(nearSpace, { 
					x = i + math.random(1, 612) - 100, 
					y = j + math.random(1, 612) - 100, 
					scale = 0.2,
					image = SpaceImages.StarImages[math.random(1, table.getn(SpaceImages.StarImages))], 
				}) 
			end
		end
	end
	
	for _, tile in ipairs(addBuffer(nearSpace, size:scale(Constant.Space.QuarterScale))) do
		table.insert(nearSpace, tile)
	end
	
	local closeSpace = { }
	
	for i = 1, size.x * Constant.Space.HalfScale, 512 do
		for j = 1, size.y * Constant.Space.HalfScale, 512 do
			while math.random() > 0.8 do 
				table.insert(closeSpace, { 
					x = i + math.random(1, 612) - 100, 
					y = j + math.random(1, 612) - 100, 
					scale = .4, --math.random(0.4, 0.6),
					image = SpaceImages.StarImages[math.random(1, table.getn(SpaceImages.StarImages))], 
				}) 
			end
		end
	end
	
	for _, tile in ipairs(addBuffer(closeSpace, size:scale(Constant.Space.HalfScale))) do
		table.insert(closeSpace, tile)
	end
	
	local space = {
		output = output,
		x = size.x,
		y = size.y,
		distantSpace = distantSpace,
		nearSpace = nearSpace,
		closeSpace = closeSpace,
		distantCamera = Camera.new(Vector2d.new(0, 0)),
		nearCamera = Camera.new(Vector2d.new(0, 0)),
		closeCamera = Camera.new(Vector2d.new(0, 0)),
		position = { x = 0, y = 0 },
	}

	return space
end

function load()
	table.insert(SpaceImages.Orientation, { theta = 0, ox = 0, oy = 0 })
	table.insert(SpaceImages.Orientation, { theta = math.pi / 2, ox = -512, oy = 0 })
	table.insert(SpaceImages.Orientation, { theta = math.pi, ox = -512, oy = 512 })
	table.insert(SpaceImages.Orientation, { theta = 3 * math.pi / 2, ox = 0, oy = 512 })
end

-- Updates the state of the world space.
-- @param
-- @param position Vector2d describing the position.
-- @param cameraMode Current camera mode.
function update(self, position)
	if not position then 
		self.output('Returning from Space:update because dx or dy is nil.')
		return 
	end

	self.distantCamera:moveAbsolute(position:scale(Constant.Space.EighthScale))
	self.nearCamera:moveAbsolute(position:scale(Constant.Space.QuarterScale))
	self.closeCamera:moveAbsolute(position:scale(Constant.Space.HalfScale))
end

-- TODO: could optimize with a sorted list of x and y positions so the 
--       loop is much smaller. These tiles are only inserted once and
--       then they never change.
function draw(self)
	local cameraPosition
	local height, width = love.graphics.getHeight(), love.graphics.getWidth()
	self.distantCamera:attach()
	cameraPosition = self.distantCamera:getPosition()
	for _, tile in ipairs(self.distantSpace) do
		if tile.x - width <= cameraPosition.x and tile.x + width >= cameraPosition.x and 
			tile.y - height <= cameraPosition.y and tile.y + height >= cameraPosition.y then
			love.graphics.draw(tile.image, tile.x, tile.y, tile.rotation, tile.scale, tile.scale, 256, 256)
		end
	end
	self.distantCamera:detach()

	self.nearCamera:attach()
	cameraPosition = self.nearCamera:getPosition()
	for _, tile in ipairs(self.nearSpace) do
		if tile.x - width <= cameraPosition.x and tile.x + width >= cameraPosition.x 
			and tile.y - height <= cameraPosition.y and tile.y + height >= cameraPosition.y then
			love.graphics.draw(tile.image, tile.x, tile.y, tile.rotation, tile.scale, tile.scale, 256, 256)
		end
	end	
	self.nearCamera:detach()
	
	love.graphics.setBlendMode('additive')
	
	self.closeCamera:attach()
	cameraPosition = self.closeCamera:getPosition()
	for _, tile in ipairs(self.closeSpace) do
		if tile.x - width <= cameraPosition.x and tile.x + width >= cameraPosition.x 
		and tile.y - height <= cameraPosition.y and tile.y + height >= cameraPosition.y then
			love.graphics.draw(tile.image, tile.x, tile.y, 0, tile.scale, tile.scale)
		end
	end	
	self.closeCamera:detach()
	
	love.graphics.setBlendMode('alpha')
end

