--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

module(..., package.seeall)
require'Class'
require'Vector2d'

local ThingMetatable = Class(Thing)

-- Creates a new thing.
-- @drawFunction Function used to draw the thing.
-- @position Vector2d describing the thing's position.
-- @velocity Vector2d describing the thing's velocity.
-- @angle Angle of orientation in radians.
-- @range Maximum range the thing should travel or nil if it should go indefinitely.
function new(name, drawFunction, position, velocity, angle, boundingShape, range)
	local thing = {
		name = name or 'thing',
		parent = nil,
		drawFunction = drawFunction or function() end,
		position = position or Vector2d.new(0, 0),
		velocity = velocity or Vector2d.new(0, 0),
		angle = angle or 0,
		boundingShape = boundingShape,
		range = range,
		dead = false,
		deathFunction = nil,
	}

	setmetatable(thing, ThingMetatable)
	return thing
end

-- Loads any resources the thing requires.
function load()

end

-- Updates the thing.
-- @param self The thing.
-- @param dt Delta time since the last update.
-- @return true if updated, false if should be removed.
function update(self, dt)
	if self.dead or (self.range and self.range <= 0) then
		if self.boundingShape then
			main.colliderLookup[self.boundingShape] = nil
			main.collider:remove(self.boundingShape)
		end
		return false
	else
		local distance = self.velocity:scale(dt)
		self.position = self.position:add(distance)
		main.wrap:ensureWrap(self.position)
		self.boundingShape:moveTo(self.position.x, self.position.y)
		if self.range then self.range = self.range - distance:length() end
	end
	
	return true
end

function setDeathFunction(self, deathFunction)
	self.deathFunction = deathFunction
end

-- Draws the thing.
-- @param self The thing.
function draw(self)
	main.wrap:drawObject(function(position) self.drawFunction(position, self.angle) end, self.position)
	--self.boundingShape:draw('line')
end

function die(self)
	self.dead = true
	
	print(self.deathFunction)
	if self.deathFunction then
		self.deathFunction(self.position, self.velocity)
	end
end


