--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
]]

module(..., package.seeall)
require'Class'
require'Vector2d'

local WrappedSpaceMetatable = Class(WrappedSpace)

DefaultOrigin = Vector2d.new(0, 0)

-- Creates a new wrapped space.
-- @param origin The origin for the wrapped space, including buffer.
-- @param size Vector2d describing the size of the wrapped space inside 
--             the buffer
-- @param buffer The buffer of background around the outside of the 
--               wrapped space.
function WrappedSpace.new(origin, size, buffer)
	local drawOrigin = origin:add(Vector2d.new(buffer, buffer))
	local drawEnd = drawOrigin:add(size)

	wrappedSpace = {
		origin = origin:copy(),
		size = size,
		buffer = buffer,
		drawOrigin = drawOrigin,
		drawEnd = drawEnd,
	}
	
	setmetatable(wrappedSpace, WrappedSpaceMetatable)
	return wrappedSpace
end

-- TODO: this doesn't work. Need to figure out why. Guess I don't care that much.
-- Draws the border along the wrap line for debugging purposes.
-- @param self The WrappedSpace.
function draw(self)
	main.console:writeMessage('drawing wrap border [' .. self.drawOrigin:toString() .. ']' .. self.size.x .. ', ' .. self.size.y)
	love.graphics.rectangle('line', self.drawOrigin.x, self.drawOrigin.y, self.size.x, self.size.y) 
end

-- Updates a given position if it moves out of the wrapped space and into
-- the buffer.
-- @param self The WrappedSpace.
-- @param position Vector2d describing the position.
function ensureWrap(self, position) 
	if position.x < self.drawOrigin.x then 
		position.x = self.drawEnd.x - (self.drawOrigin.x - position.x) 
	end
	if position.x > (self.drawEnd.x) then 
		position.x = self.drawOrigin.x + (position.x - self.drawEnd.x)
	end
	if position.y < self.drawOrigin.y then 
		position.y = self.drawEnd.y - (self.drawOrigin.y - position.y) 
	end
	if position.y > self.drawEnd.y then 
		position.y = self.drawOrigin.y + (position.y - self.drawEnd.y)
	end
end

-- Draws an object at the specified position using the function provided.
-- The function exists because when we are wrapping space around, we need
-- a buffer, but because we still want to display things in the buffer we
-- need to create a fake copy of it and draw it everywhere it should be 
-- seen.
-- @param self The WrappedSpace.
-- @param drawFunction Function used to draw the object in all of its positions.
-- @param position Vector2d describing the position.
function drawObject(self, drawFunction, position)
	drawFunction(position)
	
	local ghostX, ghostY = nil
	if position.x < self.drawOrigin.x + self.buffer then
		ghostX = position.x + self.size.x
	elseif position.x > self.drawEnd.x - self.buffer then
		ghostX = position.x - self.size.x
	end
	if position.y < self.drawOrigin.y + self.buffer then
		ghostY = position.y + self.size.y
	elseif position.y > self.drawEnd.y - self.buffer then
		ghostY = position.y - self.size.y
	end
	
	if ghostX and ghostY then
		drawFunction(Vector2d.new(ghostX, ghostY))
	end
	if ghostX then
		drawFunction(Vector2d.new(ghostX, position.y))
	end
	if ghostY then
		drawFunction(Vector2d.new(position.x, ghostY))
	end
end
