--[[
Copyright (c) 2012 John Glover

Permission is hereby granted, free of charge, to any person obtaining a copy
of newinst software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above copyright notice and newinst permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

TODO:
- add collision detection
- better fuel gauge
]]

require'Controllable'
require'Console'
require'Space'
require'Vector2d'
require'Input'
require'Hud'
require'Camera'
require'Resources'
require'WrappedSpace'
Constant = require'Constant'
HC = require'HardonCollider'
local anim8 = require 'anim8'

main = { }

main.console = Console.new(15, 15, love.graphics.getWidth() - 10, love.graphics.getHeight() / 4, 15)
main.cameraMode = Constant.Camera.Locked
main.cameraPosition = Vector2d.new(0, 0)

main.wrap = { }

main.collider = { }
main.colliderLookup = { }

main.entities = { }
main.animations = { }

main.input = Input.new()

main.pause = false
main.lastSpawn = 0
main.timescale = 1
main.score = 0

function handleCollision(dt, shape_a, shape_b, mtv_x, mtv_y)
	entity1, entity2 = main.colliderLookup[shape_a], main.colliderLookup[shape_b]
	if (entity1.name == Constant.Thing.Bullet and entity2.name == Constant.Thing.GoldDisc)
		or (entity1.name == Constant.Thing.GoldDisc and entity2.name == Constant.Thing.Bullet) then
		entity1:die()
		entity2:die()
	end
	
	--main.console:writeMessage('collision ' .. dt .. ', ' .. mtv_x .. ', ' .. mtv_y)
end

function endCollision(dt, shape_a, shape_b)
	--main.console:writeMessage(main.colliderLookup[shape_a].name .. ' stopped colliding with ' .. main.colliderLookup[shape_b].name)
	--main.console:writeMessage('end collision ' .. dt)
end

function love.load()
	main.lastSpawn = love.timer.getTime()
	main.anim8grid = anim8.newGrid(64, 64, 256, 256, 0, 0, 0)
	main.anim8frames = { '1-4,1', '1-4,2', '1-4,3', '1-4,4' }

	main.collider = HC(100, handleCollision, endCollision)
	
	main.size = Vector2d.new(16384, 16384)
	main.wrap = WrappedSpace.new(WrappedSpace.DefaultOrigin, main.size, 8192)
	
	Space.load()
	Hud.load()
	-- constructor takes a table with an x and y value describing space size
	main.space = Space.new(main.size, function (message) main.console:writeMessage(os.date('%X') .. ': ' .. message) end)
	main.hud = Hud.new(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
	
	main.player = Controllable.new(main.size.x, main.size.y, 150, 350, 0.1)
	main.player.output = function (message) main.console:writeMessage(os.date('%X') .. ': ' .. message) end
	main.player:load()
	love.graphics.setBackgroundColor(0, 0, 0)
	main.cameraPosition = main.size:copy()
	main.camera = Camera.new(main.size)
	
	main.input:bindKeystroke(' ', main.player.startAcceleration, main.player.stopAcceleration, main.player)
	main.input:bindKeystroke('f', nil, main.toggleCameraMode)
	main.input:bindMouse('wd', nil, main.zoom, 'out')
	main.input:bindMouse('wu', nil, main.zoom, 'in')
	main.input:bindMouse('l', main.player.startShooting, main.player.stopShooting, main.player)
	main.input:bindKeystroke('`', function( ) main.console:toggle() end, nil)
	main.input:bindKeystroke('esc', love.event.quit, nil)
	main.input:bindKeystroke('pause', function( ) main.pause = not main.pause end, nil)
	main.input:bindKeystroke('1', function( ) main.timescale = 0.5 end, nil)
	main.input:bindKeystroke('5', function( ) main.timescale = 1 end, nil)
	main.input:bindKeystroke('0', function( ) main.timescale = 1.5 end, nil)
	
	-- TODO: test code 
	main.spawnOthers( 0, 0, main.size.x + 8192, main.size.y + 8192, 500 )
end
		
function love.update(dt)
	if main.pause then return end
	dt = dt * main.timescale
	
	if main.lastSpawn + 10 < love.timer.getTime() then
		main.lastSpawn = love.timer.getTime()
		main.spawnOthers(0, 0, main.size.x + 8192, main.size.y + 8192, 10)
	end

	local mousePosition = main.camera:mousePosition()
	main.player:update(mousePosition, dt)

	if main.cameraMode == Constant.Camera.Locked then
		-- TODO: copy is a workaround because I'm too dumb for references.
		main.cameraPosition = main.player.position:copy()
	elseif main.cameraMode == Constant.Camera.TransitionToLocked then
		local cameraVector = 
			Vector2d.new(main.player.position.x - main.cameraPosition.x, 
				main.player.position.y - main.cameraPosition.y)
		if (cameraVector:length() < Constant.Camera.TransitionVelocity * dt) then
			main.cameraPosition = main.player.position
			main.cameraMode = Constant.Camera.Locked
		else
			local normalized = cameraVector:normalize()
			local transitionVelocity = normalized:scale(Constant.Camera.TransitionVelocity * dt)
			main.cameraPosition = main.cameraPosition:add(transitionVelocity)
		end
	end

	main.camera:moveAbsolute(main.cameraPosition)
	main.space:update(main.cameraPosition)
	
	local asteroidAngle = love.timer.getTime() % (2 * math.pi)
	local updatedEntities = { }
	for _, entity in ipairs(main.entities) do
		if entity:update(dt) then table.insert(updatedEntities, entity) end
		if entity.name == Constant.Thing.GoldDisc then entity.angle = asteroidAngle end
	end
	main.entities = updatedEntities
	
	local updatedAnimations = { }
	for _, animation in ipairs(main.animations) do
		print('updating animation')
		if animation.startTime + animation.time >= love.timer.getTime() then
			print('drawing explosion at ' .. animation.position:toString())
			table.insert(updatedAnimations, animation)
			animation.animation:update(dt)
		end
	end
	main.animations = updatedAnimations
	
	main.collider:update(dt)
end
 
function love.draw()
	local w, h = love.graphics.getWidth(), love.graphics.getHeight()

	-- space does its own camera handling
	main.space:draw()
	
	main.camera:attach()
	main.player:draw()
	for _, entity in ipairs(main.entities) do
		entity:draw()
	end
	for _, animation in ipairs(main.animations) do
		animation.animation:draw(animation.image, animation.position.x, animation.position.y, 0, 1, 1, 32, 32)
	end
	main.camera:detach()
	
	main.console:draw()
	main.hud:draw()
	
	main.printDebug()
end

function main.spawnOthers(x, y, width, height, count) 
	local scale = 0.25
	local x_offset = Resources.Images.Objects.GoldDisc:getWidth() / 2
	local y_offset = Resources.Images.Objects.GoldDisc:getHeight() / 2

	for i = 1, count do
		local position = Vector2d.new(math.random(x, x + width), math.random(y, y + height))
		local boundingShape = main.collider:addCircle(position.x, position.y, 16)
	
		local asteroid = Thing.new(
			Constant.Thing.GoldDisc,
			function(position, angle)
				love.graphics.draw(Resources.Images.Objects.GoldDisc, position.x, position.y, angle, scale, scale, x_offset, y_offset)
			end,
			position,
			Vector2d.new(math.random(0, 80) - 40, math.random(0, 80) - 40),
			nil,
			boundingShape, 
			nil
		)
		asteroid:setDeathFunction( function(position, velocity)
			table.insert(main.animations, 
				{
					position = position,
					velocity = velocity,
					animation = anim8.newAnimation('loop', main.anim8grid(unpack(main.anim8frames)), 0.1),
					image = Resources.Images.Effects.Explosion,
					startTime = love.timer.getTime(),
					time = 1.6,
				})
			main.score = main.score + 1
			end
		)
		
		main.colliderLookup[boundingShape] = asteroid
		
		table.insert(main.entities, asteroid)
	end
end

function main.toggleCameraMode() 		
		if main.cameraMode == Constant.Camera.Locked then
			main.cameraMode = Constant.Camera.Free
		elseif main.cameraMode == Constant.Camera.Free then
			main.cameraMode = Constant.Camera.TransitionToLocked
		end
end

function main.printDebug ( )
	local height = love.graphics.getHeight()
	love.graphics.print('fps          [' .. love.timer.getFPS(), 10, height - 100)
	love.graphics.print(string.format('Position     [%d, %d]', main.player.position.x, main.player.position.y), 10, height - 85)
	love.graphics.print('Mouse        [' .. main.camera:mousePosition():toString() .. ']', 10, height - 70)
	love.graphics.print('Velocity     [' .. main.player.velocity:length() .. ']', 10, height - 55)
	love.graphics.print('Acceleration [' .. main.player.ship.acceleration .. ']', 10, height - 40)
	love.graphics.print('CameraMode   [' .. main.cameraMode .. ']', 10, height - 25)
end
 
function love.keypressed(key, unicode)
	main.input:keypressed(key, unicode)
end

function love.keyreleased(key, unicode)
	main.input:keyreleased(key, unicode)
end

function love.mousepressed(x, y, button)
	main.input:mousepressed(x, y, button)
end

function love.mousereleased(x, y, button)
	main.input:mousereleased(x, y, button)
end

main.zoomLevels = { 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2 }
main.currentZoom = 4
function main.zoom(direction)
	main.console:writeMessage('Zooming ' .. direction .. '. Current zoom is ' .. main.camera.zoom)

	if direction == 'in' then
		main.currentZoom = math.min(table.getn(main.zoomLevels), main.currentZoom + 1)
	elseif direction == 'out' then
		main.currentZoom = math.max(1, main.currentZoom - 1)
	end
	
	main.camera.zoom = main.zoomLevels[main.currentZoom]
	main.console:writeMessage('Zoomed ' .. direction .. '. Current zoom is ' .. main.camera.zoom)
end
